(* A strict functional language with integers, first-order functions,
   and exceptions * sestoft@dina.kvl.dk 2001-03-14, 2003-03-13

   Exceptions in the functional object language are modelled without
   using exceptions in the meta-language.  Instead the interpreter
   uses continuations to represent the rest of the computation.
   The interpreters are written in continuation-passing style, and
   implement tail calls in constant space.

   This file contains two interpreters:

     * coEval1 allows exceptions to be thrown, but not handled; this
       is implemented using a single continuation (which is used by
       ordinary computations but ignored when an exception is thrown)

     * coEval2 allows exceptions to be thrown and handled; this is
       implemented using two continuations: the success continuation
       is used by ordinary computations, and the error continuation
       is used when an exception is thrown.
*)

app load ["Env"];

open Env;

datatype exn = 
    Exn of string

datatype expr = 
    CstI of int
  | CstB of bool
  | Var of string
  | Let of string * expr * expr
  | Prim of string * expr * expr
  | If of expr * expr * expr 
  | Letfun of string * string * expr * expr     (* (f, x, fbody, ebody) *)
  | Call of string * expr
  | Raise of exn
  | Handle of expr * exn * expr                 (* e1 handle exn => e2  *)

datatype value = 
    Int of int
  | RClo of string * string * expr * vfenv      (* (f, x, fbody, bodyenv) *)
withtype vfenv = value env;

datatype answer = 
    Success of int
  | Failure of string


(* This interpreter coEval1 takes the following arguments:

    * An expression e to evaluate.
    * An environment env in which to evalute it.
    * A success continuation cont which accepts as argument the value
      of the expression.

   It returns an answer: Success i or Failure s.  When the evaluation
   of e succeeds, it applies the success continuation to its result,
   and when e raises an exception (Exn s), it returns Failure s.
   Since there is no error continuation, there is no provision for
   handling raised exceptions.  
*)

local
fun coEval1 (e : expr) (env : vfenv) (cont : int -> answer) : answer =
    case e of 
        CstI i => cont i
      | CstB b => cont (if b then 1 else 0)
      | Var x  => (case lookup env x of 
                       Int i => cont i 
                     | _     => Failure "coEval1 Var")
      | Let(x, erhs, ebody) => 
            coEval1 erhs env (fn xval => 
			      let val env1 = bind1 env (x, Int xval)
			      in coEval1 ebody env1 cont end)
      | Prim(ope, e1, e2) => 
            coEval1 e1 env 
            (fn i1 =>
             coEval1 e2 env 
             (fn i2 =>
              case ope of
                  "*" => cont(i1 * i2)
                | "+" => cont(i1 + i2)
                | "-" => cont(i1 - i2)
                | "=" => cont(if i1 = i2 then 1 else 0)
                | "<" => cont(if i1 < i2 then 1 else 0)
                | _   => Failure "unknown primitive"))
      | Letfun(f, x, fbody, ebody) => 
            let val env1 = bind1 env (f, RClo(f, x, fbody, env))
            in coEval1 ebody env1 cont end
      | Call(f, earg) => 
	    let val fclosure = lookup env f
	    in 
		case fclosure of
		    RClo (f, x, fbody, fenv) =>
			coEval1 earg env  
			        (fn argv =>
				 let 
				     val env2 = bind1 fenv (f, fclosure)
				     val env3 = bind1 env2 (x, Int argv)
				 in coEval1 fbody env3 cont end)
		  | _ => Failure "eval Call: not a function"
	    end
      | If(e1, e2, e3) => 
	    coEval1 e1 env
	            (fn b => if b<>0 then coEval1 e2 env cont
			     else coEval1 e3 env cont)
      | Raise (Exn s) => Failure s
      | Handle(e1, exn, e2) => 
             Failure "Not implemented"
in
    fun eval1 e env = coEval1 e env (fn v => Success v)
    fun run1 e = eval1 e Env.empty
end


(* This interpreter coEval2 takes the following arguments:

    * An expression e to evaluate.
    * An environment env in which to evalute it.
    * A success continuation cont which accepts as argument the value
      of the expression.
    * A error continuation econt, which is applied when an exception
      is thrown

   It returns an answer: Success i or Failure s.  When the evaluation
   of e succeeds, it applies the success continuation to its result,
   and when e raises an exception exn, it applies the failure
   continuation to exn.  The failure continuation may choose to handle
   the exception.  
*)

local
fun coEval2 (e : expr) (env : vfenv) 
            (cont : int -> answer) (econt : exn -> answer) : answer =
    case e of 
        CstI i => cont i
      | CstB b => cont (if b then 1 else 0)
      | Var x  => (case lookup env x of 
                       Int i => cont i 
                     | _     => Failure "coEval2 Var")
      | Let(x, erhs, ebody) => 
            coEval2 erhs env (fn xval => 
			      let val env1 = bind1 env (x, Int xval)
			      in coEval2 ebody env1 cont econt end)
                             econt
      | Prim(ope, e1, e2) => 
            coEval2 e1 env 
            (fn i1 =>
             coEval2 e2 env 
             (fn i2 =>
              case ope of
                  "*" => cont(i1 * i2)
                | "+" => cont(i1 + i2)
                | "-" => cont(i1 - i2)
                | "=" => cont(if i1 = i2 then 1 else 0)
                | "<" => cont(if i1 < i2 then 1 else 0)
                | _   => Failure "unknown primitive") econt) econt
      | Letfun(f, x, fbody, ebody) => 
            let val env1 = bind1 env (f, RClo(f, x, fbody, env))
            in coEval2 ebody env1 cont econt end
      | Call(f, earg) => 
	    let val fclosure = lookup env f
	    in 
		case fclosure of
		    RClo (f, x, fbody, fenv) =>
			coEval2 earg env  
			        (fn argv =>
				 let 
				     val env2 = bind1 fenv (f, fclosure)
				     val env3 = bind1 env2 (x, Int argv)
				 in coEval2 fbody env3 cont econt end)
				econt
		  | _ => raise Fail "eval Call: not a function"
	    end
      | If(e1, e2, e3) => 
            coEval2 e1 env (fn b =>
                          if b<>0 then coEval2 e2 env cont econt
                          else coEval2 e3 env cont econt) econt
      | Raise exn => econt exn
      | Handle(e1, exn, e2) => 
             let fun econt1 exn1 =
                     if exn1 = exn then coEval2 e2 env cont econt
                                   else econt exn1
             in coEval2 e1 env cont econt1 end

in
    (* The top-level error continuation returns the continuation, 
       adding the text Uncaught exception *)

    fun eval2 e env = 
        coEval2 e env 
               (fn v => Success v) 
               (fn (Exn s) => Failure ("Uncaught exception: " ^ s))
    fun run2 e = eval2 e Env.empty
end


(* Examples in abstract syntax *)

val ex1 = Letfun("f1", "x", Prim("+", Var "x", CstI 1), 
                 Call("f1", CstI 12));

(* Factorial *)

val ex2 = Letfun("fac", "x", 
                 If(Prim("=", Var "x", CstI 0),
                    CstI 1,
                    Prim("*", Var "x", 
                               Call("fac", Prim("-", Var "x", CstI 1)))),
                 Call("fac", Var "n"));

val fac10 = eval1 ex2 (Env.fromList [("n", Int 10)]);

(* Example: deep recursion to check for constant-space tail recursion *)

val exdeep = Letfun("deep", "x", 
                    If(Prim("=", Var "x", CstI 0),
                       CstI 1,
                       Call("deep", Prim("-", Var "x", CstI 1))),
                    Call("deep", Var "n"));
    
fun rundeep n = eval1 exdeep (Env.fromList [("n", Int n)]);

(* Example: throw an exception inside expression *)

val ex3 = Prim("*", CstI 11, Raise (Exn "outahere"));

(* Example: throw an exception and handle it *)

val ex4 = Handle(Prim("*", CstI 11, Raise (Exn "Outahere")),
                 Exn "Outahere", 
		 CstI 999);

(* Example: throw an exception in a called function *)

val ex5 = 
    Letfun("fac", "x", 
           If(Prim("<", Var "x", CstI 0),
              Raise (Exn "negative x in fac"),
              If(Prim("<", Var "x", CstI 0),
                 CstI 1,
                 Prim("*", Var "x", 
                           Call("fac", Prim("-", Var "x", CstI 1))))),
           Call("fac", CstI ~10));

(* Example: throw an exception but don't handle it *)

val ex6 = Handle(Prim("*", CstI 11, Raise (Exn "Outahere")),
                 Exn "Noway", 
		 CstI 999);

val ex7 = Handle(Prim("*", CstI 11, Raise (Exn "Outahere")),
                 Exn "Outahere", 
		 CstI 999);

